package com.lightsky.monitor

import com.lightsky.utils.ContextUtils
import com.lightsky.utils.LogUtils
import com.lightsky.utils.ProcessUtils
import java.text.SimpleDateFormat

/**
 * Created by xupengpai on 2021/1/5.
 */
class AppMonitor {

    companion object{

        val TAG = "AppMonitor"
        @JvmField
        val blockInfoMap = mutableMapOf<String, BlockInfo>()
        var onlySimpleLog = true
        var enableLog = true
        @JvmField
        var isStatReported = false

        @JvmField
        var TIME_BEGIN = 1
        @JvmField
        var TIME_END = 0


//        fun begin(type:String, key: String){
//            val blockInfo: BlockInfo = if (blockInfoMap.containsKey(key)) blockInfoMap[key]!! else BlockInfo()
//            if (blockInfo.beginTime != null){
//                //同一个key未结束重复调用开始，可能是外部调用逻辑问题
//                LogUtils.e(TAG, "重复调用begin()，请确认逻辑是否正确,,[${key}]\"")
//            }
//            blockInfo.type = type
//            blockInfo.key = key
//            blockInfo.beginTime = System.currentTimeMillis()
//            blockInfoMap[key] = blockInfo
//
//        }


        /**
         * 直接记录锚点，约等于begin、end同时调用
         */
        @JvmStatic
        @JvmOverloads
        fun tap(key: String, override: Boolean = true): BlockInfo?{
            return begin(key, override)?.also {
                it.endTime = it.beginTime
            }
        }

        var isMainprocess: Boolean? = null

        fun isMainProcess(): Boolean{
            return isMainprocess ?: let {
                isMainprocess = ProcessUtils.getCurrentProcessName().equals(ContextUtils.getApplicationContext().packageName)
                isMainprocess!!
            }
        }

        /**
         * 开始一个执行监控，记录一个执行锚点并收集相关信息
         */
        @JvmStatic
        @JvmOverloads
        fun begin(key: String, override: Boolean = false): BlockInfo?{

            try {
                //忽略非主进程
                if(!isMainProcess())
                    return null
                val blockInfo: BlockInfo = if (blockInfoMap.containsKey(key)) blockInfoMap[key]!! else BlockInfo()
                if (blockInfo.beginTime != 0L && !override) {
                    //同一个key未结束重复调用开始，可能是外部调用逻辑问题
                    LogUtils.e(TAG, "已经存在锚点，请确认逻辑是否正确,[${key}]")
                    return blockInfo
                }
                blockInfo.key = key
                blockInfo.beginTime = System.currentTimeMillis()
                blockInfoMap[key] = blockInfo

                if (enableLog && !onlySimpleLog) {
                    LogUtils.e(TAG, "[${key}] begin at ${blockInfo.beginTime} ${formatTime(blockInfo.beginTime)}")
                }
                return blockInfo
            }catch (e: Exception){
                LogUtils.e(TAG, "", e)
            }
            return null

        }

        fun formatTime(time: Long): String{
            return SimpleDateFormat("yyyy-MM-dd hh:mm:ss.SSS").format(time)
        }

        /**
         * 结束一个执行监控同时记录一个执行锚点
         */
        @JvmStatic
        fun end(key: String): BlockInfo?{
            try {
                //忽略非主进程
                if(!isMainProcess())
                    return null
                val blockInfo: BlockInfo? = blockInfoMap[key]
                if (blockInfo == null){
                    //没有开始，逻辑有误
                    LogUtils.e(TAG, "请先调用begin(),[${key}]")
                    return null
                }
                blockInfo.endTime = System.currentTimeMillis()
                if(enableLog && !onlySimpleLog) {
                    LogUtils.e(TAG, "[${key}] end at ${blockInfo.endTime} ${formatTime(blockInfo.beginTime)}")
                }
                if(enableLog) {
                    LogUtils.e(TAG, "[${key}] time: ${(blockInfo.endTime - blockInfo.beginTime)} ms")
                }
                return blockInfo
            }catch (e: Exception){
                LogUtils.e(TAG, "", e)
            }
            return null
        }

        @JvmStatic
        fun getBlock(key: String): BlockInfo?{
            return blockInfoMap[key]
        }

        @JvmStatic
        fun getBeginTime(key: String): Long{
            return blockInfoMap[key]?.beginTime ?: 0
        }

        @JvmStatic
        fun getEndTime(key: String): Long{
            return blockInfoMap[key]?.endTime ?: 0
        }


        /**
         * 获取一个区间的耗时
         * beginKey 开始块
         * beginOfBeginBlock 是否从开始块的开始算起，为true则使用beginTime计算，为false则使用endTime计算
         * endKey 结束块
         * beginOfEndBlock 是否从结束块的开始算起，为true则使用beginTime计算，为false则使用endTime计算
         */
        @JvmStatic
        fun getTime(beginKey: String, timeOfBeginBlock: Int, endKey: String, timeOfEndBlock: Int): Long{
            try {
                val beginTime = if (timeOfBeginBlock == TIME_BEGIN) getBeginTime(beginKey) else getEndTime(beginKey)
                val endTime = if (timeOfEndBlock == TIME_BEGIN) getBeginTime(endKey) else getEndTime(endKey)

                if (beginTime <= 0L) {
                    LogUtils.e(TAG, "开始块时间异常,[${beginKey},${timeOfBeginBlock}]")
                    return -1
                }
                if (endTime == 0L) {
                    LogUtils.e(TAG, "结束块时间异常,[${endKey},${timeOfEndBlock}]")
                    return -1
                }
                return endTime - beginTime
            }catch (e: Exception){
                LogUtils.e(TAG, "", e)
                return -1
            }
        }

        fun getBlockTimeTag(timeOfBlock: Int): String{
            return when(timeOfBlock){
                TIME_BEGIN -> "begin"
                else -> "end"
            }
        }

        /**
         * 按照指定锚点来统计和输出执行时间
         * beginKey 开始锚点
         * timeOfBeginBlock 指定使用开始锚点的开始时间还是结束时间
         * endKey 结束锚点
         * timeOfEndBlock 指定使用结束锚点的开始时间还是结束时间
         */
        @JvmStatic
        fun log(beginKey: String, timeOfBeginBlock: Int, endKey: String, timeOfEndBlock: Int){
            try {
                //忽略非主进程
                if(!isMainProcess())
                    return
                LogUtils.e(TAG, makeFormatLog(beginKey, timeOfBeginBlock, endKey, timeOfEndBlock))
            }catch (e: Exception){
                LogUtils.e(TAG, "", e)
            }
        }

        @JvmStatic
        fun makeFormatLog(beginKey: String, timeOfBeginBlock: Int, endKey: String, timeOfEndBlock: Int): String{
            //忽略非主进程
            if(!isMainProcess())
                return ""
            val time = getTime(beginKey, timeOfBeginBlock, endKey, timeOfEndBlock)
            return "[${beginKey} ${getBlockTimeTag(timeOfBeginBlock)} -> ${endKey} ${getBlockTimeTag(timeOfEndBlock)}] time ${time} ms"
        }

        /**
         * 代理调用某函数或者代码块，主要用于简化统计代码
         * PS：可能会导致堆栈中多出一层函数调用
         */
        @JvmStatic
        fun call(key: String, call: IMethodCall){
            begin(key)
            call.invoke()
            end(key)
        }

    }

    open class BlockInfo{
        var type: String? = null
        var key: String? = null
        var beginTime: Long = 0
        var endTime: Long = 0
    }

}